!--------------------------------!
! Module: omflatfield_mod
!
! Authors: Phil Williams - MSSL
!
! Description: Creates and applies an OM flatfield to an OSW image.
!
! Comments:
!
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!
! History : Created  : March 98 - PSW
!           Modified : 13 May 99 - Spring clean of code.
!                                - Remove calls to CAL.
!                                - Upgrade shift and add to include incremental shifts.
!	    Modified - May 2000 - JES - code to cope with dummy trackkinh history file
!           modified - 2nd June 2000 - AMK - Added code to tidy up allocated memory
!
!          Modified 25th July 2000 - CB - bug removed in which tmp_tracking
!          was being opened twice for reading.
!
!          Modified 23nd August 2000 - AMK - Temporary fix in tmp_image_2.
!          This sets a pixel value that is flagged as missing to zero. If this is not
!          done then OMDETECT will fail.
!          The long term solution is to flag the quality array and modify OMDETECT
!          so that the quality array is processed.
!
!          Modified 4th October 2000 (CB) - Minor code changes made to reduce
!          the execution time. 
!          
!          Modified 17th October 2000 (AMK) - Bug fix
!          16.11.2000 Modified to trap pixel values with values less than
!          the value -blank + 1 or greater than the value blank - 1 
!          Modified 19th March 2001 (CB) - Checks added to stop array subscripts
!          going out of range.        
!          Modified 5th October 2001 - Check added to see if the in-orbit
!          flatfield is an artifical one (all pixels equal to 1)- if so
!          no flatfielding carried out.
!**************************************************************************   
module omflatfield_mod

  use param
  use dal
  use sasdal
  use errorHandling
  use readmodes_mod
  use copytable_mod
  use copyimage_mod

  implicit none

  private

  public omflatfield

    !***********************************************
    ! Want the following to be global variables
    ! for access by various subroutines
    !***********************************************
    real(kind=single)                               :: realBinX       ! real version of binX
    real(kind=single)                               :: realBinY       ! real version of binY
    real(kind=single)                               :: binXTimesBinY  ! real(binX x binY)
    real(kind=single), dimension(:, :), pointer :: inOrbitFlatP       ! In orbit flatfield pointer.
  
    real(kind=single), dimension(:, :), pointer     :: outOSWP        ! Output observation science window pointer.
    real(kind=single), dimension(:, :), allocatable :: outOSW         ! Output observation science window pointer.
    real(kind=single), allocatable, dimension(:, :) :: shiftedFlat    ! Tracking shifted flatfield.
    real(kind=single)                               :: missingPixelValue ! Changed from int32 on 15.5.2001 and from blank
    real(kind=single)                               :: realXMin         ! real version of xMin
    real(kind=single)                               :: realYMin         ! real version of yMin
    integer(kind=int8), dimension(:, :), pointer    :: qualArrayP        ! Quality array pointer.
    integer(kind=int32) :: sampleFactor                               ! Sampling factor.
    real(kind=single)                               :: realSampleFactor ! real version of sampleFactor
    

    type trackingFrame
       real(kind=single) :: dX                                        ! Delta x.
       real(kind=single) :: dY                                        ! Delta y.
       real(kind=single) :: quality                                   ! Quality.
    end type trackingFrame
    type(trackingFrame), allocatable, dimension(:) :: tf              ! Tracking frame array.


contains

  !---------------------------------------------------------------------
  subroutine omflatfield
    !---------------------------------------------------------------------
    character(len=8) :: dataMode                                      ! Input file data mode.
    character(len=255) :: set                                         ! OM Image filename.
    character(len=255) :: inOrbitFlatSet                             

    character(len=255) :: thxSet                                      ! OM Tracking History Data Auxiliary filename.
    character(len=255) :: PPSFlatSet                                  ! OM PPS Flatfield filename.
    character(len=255) :: outSet                                      ! OM Flatfielded Image filename.

    type(DataSetT) :: inDataSetH                                      ! Input dataset handle.
    type(DataSetT) :: thxDataSetH                                     ! Tracking History Data Auxiliary dataset handle.
    type(DataSetT) :: inOrbitFlatDataSetH                             ! In Orbit Flatfield dataset handle.
    type(DataSetT) :: outDataSetH                                     ! Flatfield image dataset handle.
    type(ArrayT) :: inArrayH                                          ! Input array handle.
    type(ArrayT) :: inOrbitFlatArrayH                                 ! In Orbit Flatfield array handle.
    type(TableT) :: inTableH                                          ! Input table handle.
    type(TableT) :: thxTableH                                         ! Tracking History Data Auxiliary table handle.
    type(ColumnT), dimension(4) :: inColumnH                          ! Input image column handle.
    type(ColumnT), dimension(3) :: thxColumnH                         ! Tracking History Data Auxiliary column handle.
  
    integer(kind=int16) :: nFrames                                    ! Number of tracking frames.

    integer(kind=int32), dimension(:), pointer :: windowX0P           ! WindowX0 pointer.
    integer(kind=int32), dimension(:), pointer :: windowY0P           ! WindowY0 pointer.
    integer(kind=int32), dimension(:), pointer :: windowDXP           ! WindowDX pointer.
    integer(kind=int32), dimension(:), pointer :: windowDYP           ! WindowDY pointer.
    integer(kind=int32), dimension(:), pointer :: dXP                 ! DX pointer.
    integer(kind=int32), dimension(:), pointer :: dYP                 ! DY pointer.
    integer(kind=int32), dimension(:), pointer :: qualityP            ! Quality pointer.
    integer(kind=int32) :: binAx1                                     ! DPU x-axis binning factors.
    integer(kind=int32) :: binAx2                                     ! DPU y-axis binning factors.
    integer(kind=int32) :: binX                                       ! Bin x.
    integer(kind=int32) :: binY                                       ! Bin y.
    integer(kind=int32) :: windowX0                                   ! WindowX0.
    integer(kind=int32) :: windowY0                                   ! WindowY0.
    integer(kind=int32) :: windowDX                                   ! WindowDX.
    integer(kind=int32) :: windowDY                                   ! WindowDY.
    integer(kind=int32) :: xMin                                       ! Minimum x value for window.
    integer(kind=int32) :: xMax                                       ! Maximum x value for window.
    integer(kind=int32) :: yMin                                       ! Minimum y value for window.
    integer(kind=int32) :: yMax                                       ! Maximum y value for window.
    integer(kind=int32) :: allocErr                                   ! Allocate error status.
    integer(kind=int32) :: i                                          ! Count.
    integer(kind=int32) :: j                                          ! Count.
    integer(kind=int32) :: x                                          ! X.
    
    real(kind=single), dimension(:, :), pointer :: inOSWP             ! Input observation science window pointer.
    real(kind=single), allocatable, dimension(:, :) :: inOSW          ! Input observation science window.
  
    !*************************************************************************
    ! The following variables have been added to reduce the execution time
    !*************************************************************************
 
    integer                                         :: istat
 
    integer dim(2)
    integer ind, iMax, jMax
  
 
    type(ArrayT) :: qualArrayH                                        ! Quality Array handle.
    logical(kind=bool) :: binBPE                                      ! BPE binning flag.
    logical :: goodFlatField = .TRUE.
  
    logical(kind=bool), dimension(:, :), allocatable :: qualArray     ! Quality array.

  
    !*********************************************
    ! Read in the program command-line parameters.
    !*********************************************
    sampleFactor = intParameter("samplefactor")
    realSampleFactor = real(sampleFactor)
    set = stringParameter("set")
    thxSet = stringParameter("thxset")
    inOrbitFlatSet = stringParameter("inorbitflatset")

    PPSFlatSet = stringParameter("tsflatset")
    outSet = stringParameter("outset")

    ! Read in the OM OSW Image file.

    inDataSetH = dataSet(set, read)

    ! Determine file type from datamode in primary header.

    dataMode = stringAttribute(inDataSetH, "DATAMODE")

    if (dataMode .ne. "IMAGING") then

        write (errstr, *) "Invalid datamode ", dataMode, " in set ", set
        call fatal("badFileMode")
    end if

  
    !*****************************************************************
    !
    ! Read in the science window keywords from the primary header.
    !
    ! missingPixelValue is the value representing a missing value
    !*****************************************************************
    missingPixelValue = real32Attribute(inDataSetH, "MISSPIX")
    binBPE = booleanAttribute(inDataSetH, "BINBPE")
    binAx1 = int32Attribute(inDataSetH,   "BINAX1")
    binAx2 = int32Attribute(inDataSetH,   "BINAX2")

    ! Calculate the image binning.

    binX = 2 ** binAx1
    binY = 2 ** binAx2

    if (binBPE) then
        binX = binX * 2
        binY = binY * 2
    end if
    realBinX = real(binX)                   ! Added to speed up the program
    realBinY = real(binY)                   ! ditto
    binXTimesBiny = real(binX * binY)       ! ditto

    ! Get a handle on the modes binary table extension.

    inTableH = table(inDataSetH, "MODES")
 
    ! Get handles and pointers on the modes binary table extension &
    ! columns and extract the window data.
 
    inColumnH(1) = column(inTableH, "WINDOWX0", read)
    windowX0P => int32Data(inColumnH(1))
    windowX0 = windowX0P(0)

    inColumnH(2) = column(inTableH, "WINDOWY0", read)
    windowY0P => int32Data(inColumnH(2))
    windowY0 = windowY0P(0)

    inColumnH(3) = column(inTableH, "WINDOWDX", read)
    windowDXP => int32Data(inColumnH(3))
    windowDX = windowDXP(0)

    inColumnH(4) = column(inTableH, "WINDOWDY", read)
    windowDYP => int32Data(inColumnH(4))
    windowDY = windowDYP(0)

    xMin = windowX0
    realXMin = real(xMin)              ! Added to speed the program up
    xMax = windowX0 + windowDx - 1
    yMin = windowY0
    realYMin = real(yMin)              ! Added to speed the program up
    yMax = windowY0 + windowDy - 1


    !*****************************************
    ! Allocate space for the quality array.
    !*****************************************
    allocate (qualArray(1 : windowDX / binX, 1 : windowDY / binY), stat = allocErr)
    call CheckAllocation(allocErr, "qualArray")

    ! Get a handle and pointer on the quality array and extract the quality data.
    qualArrayH = array(inDataSetH, "QUALITY", read)
    qualArrayP => int8Array2Data(qualArrayH)

    where (qualArrayP /= 0) 
        qualArray = .true.
    elsewhere
        qualArray = .false.
    end where

    !**********************************************************
    ! Allocate space for the input observation science window.
    !**********************************************************
    allocate (inOSW(xMin : xMax, yMin : yMax), stat=allocErr)
    call CheckAllocation(allocErr, "inOSW")


    ! Get a handle and pointer on the input observation science window &
    ! array and extract the OSW data.

    inArrayH = array(inDataSetH, "PRIMARY", read)
  
    dim = dimensions(inArrayH)
  
    inOSWP => real32Array2Data(inArrayH)

    !***************************
    !Get the image dimensions
    !***************************
    dim = dimensions(inArrayH)
    imax = dim(1) - 1
    jMax = dim(2) - 1

    inOSW = 0.0
   
    do i = xMin, xMax
        x = min(imax, floor((i - xMin) / realBinX))
        do j = yMin, yMax
            ind = min(floor((j - yMin) / realBinY), jMax)
   
            inOSW(i, j) = inOSWP(x, ind) / binXTimesBinY
                   
        end do
    end do

    ! Release handles on data.
    call release(inArrayH)
    call release(inColumnH(1))
    call release(inColumnH(2))
    call release(inColumnH(3))
    call release(inColumnH(4))
    call release(inTableH)
 
    !************************************************************
    ! Get the tracking history information, if it exists
    !***********************************************************
    thxDataSetH = dataSet(thxset, read)

    ZERODRIFT: if (hasBlock(thxDataSetH, "ZERODRIFT")) then ! No tracking-history information
       write (errstr, *) "Missing OM Tracking History Data Auxiliary File. Using ZERODRIFT option"
       write (errstr,*) "No PPS Tracking History Products available"
       call warning("ZERODRIFT")  
       nFrames = 0
    else    
        ! Determine file type from datamode in primary header.
        dataMode = stringAttribute(thxDataSetH, "DATAMODE")


        if (dataMode .ne. "TRACKING") then

            ! Incompatible file.
            write (errstr, *) "Invalid datamode ", dataMode, " in set ", thxSet
            call fatal("badFileMode")
        end if
   

        ! Get a handle on the tracking history binary table extension.
        thxTableH = table(thxDataSetH, "OMTHX1")

        ! Get the number of frames in the tracking history binary table extension.
        nFrames = numberOfRows(thxTableH)

        ! Allocate space for the tracking frames.
        allocate (tf(0 : nFrames), stat=allocErr)
        call CheckAllocation(allocErr, "tf")


        ! Get handles and pointers on the tracking history binary &
        ! table extension columns and extract the tracking data.

        thxColumnH(1) = column(thxTableH, "DX", read)
        dXP => int32Data(thxColumnH(1))
        tf(0) % dX = 0.0
        tf(1 : nFrames) % dX = real(dXP / 1000.0)

        thxColumnH(2) = column(thxTableH, "DY", read)
        dYP => int32Data(thxColumnH(2))
        tf(0) % dY = 0.0
        tf(1 : nFrames) % dY = real(dYP / 1000.0)

        thxColumnH(3) = column(thxTableH, "QUALITY", read)
        qualityP => int32Data(thxColumnH(3))
        tf(0) % quality = 0.0
        tf(1 : nFrames)%quality = real(qualityP / 1000000.0)

        ! Release handles on data.
        call release(thxColumnH(1))
        call release(thxColumnH(2))
        call release(thxColumnH(3))
        call release(thxTableH)
        !*****************************************
        ! End of zerodrift block
        !*****************************************
       
    end if ZERODRIFT ! no tracking-history file check


    call release(thxDataSetH)

    !**************************************************
    ! Allocate an array to hold the output image
    !**************************************************
    xMin = windowX0
    xMax = windowX0 + windowDx - 1
    yMin = windowY0
    yMax = windowY0 + windowDy - 1
    allocate (outOSW(xMin : xMax, yMin : yMax), stat=allocErr)
    call CheckAllocation(allocErr, "outOSW")


    ! Read in the OM In Orbit Flatfield file.

    inOrbitFlatDataSetH = dataSet(inOrbitFlatSet, read)

  
    ! Get a handle and pointer on the in orbit flatfield array and extract the flatfield data.

    inOrbitFlatArrayH = array(inOrbitFlatDataSetH, "PRIMARY", read)
    inOrbitFlatP => real32Array2Data(inOrbitFlatArrayH)

    !*********************************************************
    ! Check to see if the flatfield is just a unit one-
    ! If so no point in doing the flat-field correction.
    ! goodFlatField will be TRUE if real flatfield- FALSE 
    ! otherwise
    !********************************************************
    goodFlatField = CheckFlatField(inOrbitFlatArrayH)

    if(goodFlatField) then
        !******************************************
        ! Create a shifted flatfield image
        !******************************************
  
        call CreateShiftedFlat(inOrbitFlatArrayH, nFrames, &
                     windowX0, windowDX,  windowY0, windowDY)

        where (shiftedFlat .gt. 0.0)       
            outOSW = inOSW / shiftedFlat
        elsewhere    
            outOSW = inOSW
        end where   
    else
        outOSW = inOSW
    end if

    call release(inOrbitFlatDataSetH)
 
    !***********************************
    ! Set the quality array
    !***********************************
    call SetQualityArray(xMin, xMax, yMin, yMax)


    !*********************************************
    ! Create the output OM Image file.
    !*********************************************
    outDataSetH = dataSet(outSet, Create)
    !***************************************************************    
    ! Create the flat-fielded image file
    !**************************************************************
    call CreateFlatFieldedImageFile(goodFlatField, inDataSetH, & 
                                    outDataSetH, PPSFlatSet,   &
                                    xMin, Xmax, yMin, yMax,    &
                                    missingPixelValue)
  
  
    call release(inDataSetH)
 
    deallocate(inOSW,stat = istat)
    call CheckDeallocation(istat, "inOSW")

!    deallocate(tf, stat = istat)
!    call CheckDeallocation(istat, "tf")
  
    deallocate(outOSW,stat=istat)
    call CheckDeallocation(istat, "outOSW")

    deallocate(qualArray,stat=istat)
    call CheckDeallocation(istat, "qualArray")

    if (allocated(shiftedFlat)) then
        deallocate(shiftedFlat, stat = allocErr)
        call CheckDeallocation(allocErr, "shiftedFlat")
    end if
end subroutine omflatfield
!************************************************************************
!
! Added by CB
!
! This is a temporary subroutine that has been added. It will be removed
! when OMDETECT is able to process the quality array in order to detect
! problem pixel values in the image.
!
!************************************************************************
subroutine  CheckArray(title, array, imin, imax, jmin, jmax , missingPixelValue)
    character(len=*), intent(in) :: title
    integer                       :: i, j
    integer                       :: noMissingPixels
    integer, intent(in)           :: imin, imax, jmin, jmax
    real(kind=single), intent(in) :: missingPixelValue
    real(kind = single), intent(inout), dimension(imin : imax, jmin : jmax) :: array

    !********************************************
    ! Want to trap pixels with either the value
    ! blank - 1, or the value -blank + 1
    !********************************************
 
    noMissingPixels = 0

    do i = imin, imax
       do j = jmin, jmax

          if (abs(array(i, j) - missingPixelValue) .lt. 0.5) then

             !write(*,*)'Bad pixel at location ',i,j
             write(errstr,*)'Correcting bad pixel value at ',i,j
             call message(AppMsg,Verbose)
             array(i, j) = 0.0
             noMissingPixels = noMissingPixels + 1
          end if
       end do
    end do

    if (noMissingPixels .gt. 0) then
       write(errstr, *) "Bad pixels detected in array ", title
       call message(AppMsg,Verbose)
       write(errstr, *) "No missing pixels detected = ", noMissingPixels
       call message(AppMsg,Verbose)
       write(errstr, *) "(Missing pixel value = ", missingPixelValue, ")"
       call message(AppMsg,Verbose)
    end if

end subroutine  CheckArray
!******************************************************
! Function CheckFlatField
! Checks to see if the flatfield is just a dummy one
! (ie all pixels are 1)
!*****************************************************
logical Function CheckFlatField(inOrbitFlatArrayH)
    Type(ArrayT), intent(in) :: inOrbitFlatArrayH

    character(len=30) :: origin = "UNKNOWN"
    
    !*********************************************************
    ! Check to see if the flatfield is just a unit one-
    ! If so no point in doing the flat-field correction.
    !********************************************************

    if (hasAttribute(inOrbitFlatArrayH, "ORIGIN")) then
        origin = trim(stringAttribute(inOrbitFlatArrayH, "ORIGIN"))
        if (origin .eq. "Used CAL template") then
            call warning("Uniform flatfield- no correction to image will be made")
            CheckFlatField = .FALSE.
            return
        end if
    end if
        
    if (minval(inOrbitFlatP) .eq. maxval(inOrbitFlatP)) then
        call warning("Uniform flatfield- no correction to image will be applied")
        CheckFlatField = .false.
        return 
    end if 
    CheckFlatField = .true.
end Function CheckFlatField

!***************************************************
! CheckAllocation
!***************************************************
subroutine CheckAllocation(status, string)
    integer, intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a32, a20)') "Could not allocate memory for : ", string 
       write (errstr, *) comment
       call fatal("allocErr") 
    end if

end subroutine CheckAllocation
!***************************************************
! CheckDeallocation
!***************************************************
subroutine CheckDeallocation(status, string)
    integer, intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a31, a20)') "Failed to release memory for : ", trim(string) 
       write (errstr, *) comment
       call warning("allocErr", comment) 
    end if

end subroutine CheckDeallocation
!***************************************************
! CheckDeallocation
!***************************************************
subroutine  CreateFlatFieldedImageFile(goodFlatField, inDataSetH, & 
                             outDataSetH, PPSFlatSet, &
                             xMin, Xmax, yMin, yMax, &
                             missingPixelValue)
    logical, intent(in) :: goodFlatField
    type(DataSetT), intent(in) :: inDataSetH, outDataSetH
    integer(kind=int32), intent(in) :: xMin                          ! Minimum x value for window.
    integer(kind=int32), intent(in) :: xMax                          ! Maximum x value for window.
    integer(kind=int32), intent(in) :: yMin                          ! Minimum y value for window.
    integer(kind=int32), intent(in) :: yMax                          ! Maximum y value for window.
    real(kind=single), intent(in)   :: missingPixelValue 
    character(len=*), intent(in) :: PPSFlatSet                       ! OM PPS Flatfield filename.
  
    !*****************************
    ! Local variables
    !*****************************
    type(DataSetT) :: PPSFlatDataSetH                                ! PPS flatfield dataset handle.
    type(BlockT) :: blockH                                           ! Block handle.
    type(AttributableT) :: sourceH                                   ! Source attributable handle.
    type(AttributableT) :: objectH                                   ! Object attributable handle.
    type(ArrayT) :: outArrayH                                        ! Flatfield image array handle.
    type(ArrayT) :: PPSFlatArrayH                                    ! PPS flatfield array handle.
    integer :: dim(2), iMax, jMax
    integer :: i, j
    integer(kind=int32) :: x                                         ! X.
    integer(kind=int32) :: y                                         ! Y.
    real(kind=single), dimension(:, :), pointer :: PPSFlatP          ! PPS flatfield pointer.
 

    ! Copy the input blocks and attributes to the output.

    blockH = block(inDataSetH, "PRIMARY", read)
    call copyBlock(outDataSetH, blockH)

    if (hasBlock(inDataSetH, "RAW")) then

       blockH = block(inDataSetH, "RAW", read)
       call copyBlock(outDataSetH, blockH)

    end if

    sourceH = attributable(inDataSetH)
    objectH = attributable(outDataSetH)
    call addAttributes(objectH, sourceH)

    blockH = block(inDataSetH, "MODES", read)
    call copyBlock(outDataSetH, blockH)

    blockH = block(inDataSetH, "QUALITY", read)
    call copyBlock(outDataSetH, blockH)

    ! Get a handle and pointer on the output observation science window array and insert the OSW data.

    outArrayH = array(outDataSetH, "PRIMARY", modify)
    outOSWP => real32Array2Data(outArrayH)
    dim = dimensions(outArrayH)
    iMax = dim(1) - 1
    jMax = dim(2) - 1
    

!--------------------------------------------------------------------------------------
!
! AMK - 24/08/2000
!
! This is a temporary fix until OMDETECT is able to process the quality array
!

    call CheckArray("outOSW", outOSW, xMin, Xmax, yMin, yMax, missingPixelValue)


    outOSWP = 0.0
 
    do i = xMin, xMax
        x = min(imax, floor((i - xMin) / realBinX))
        do j = yMin, yMax
            y = min(floor((j - yMin) / realBinY), jMax)
     
            outOSWP(x, y) = outOSWP(x, y) + outOSW(i, j)
              
        end do
    end do

    ! Release handles on data.

    call release(outArrayH)
    call release(outDataSetH)

    ! Create the PPS flatfield.

    PPSFlatDataSetH = dataSet(PPSFlatSet, Create)

    ! Copy some of the input blocks and attributes to the output.

    blockH = block(inDataSetH, "PRIMARY", read)
    call copyBlock(PPSFlatDataSetH, blockH)

    sourceH = attributable(inDataSetH)
    objectH = attributable(PPSFlatDataSetH)
    call addAttributes(objectH, sourceH)

    ! Update and add flat field specific attributes.

    call setAttribute(PPSFlatDataSetH, "CONTENT", "OM OSW FLAT FIELD", "Content of file")
    call setAttribute(PPSFlatDataSetH, "HDUCLASS", "OGIP", "Format conforms to OGIP/GSFC conventions")
    call setAttribute(PPSFlatDataSetH, "HDUCLAS1", "IMAGE", "File contains an image")
    call setAttribute(PPSFlatDataSetH, "HDUCLAS2", "DETMAP", "Detector map")

    ! Copy some of the input blocks and attributes to the output.

    blockH = block(inDataSetH, "MODES", read)
    call copyBlock(PPSFlatDataSetH, blockH)

    ! Get a handle and pointer on the PPS flatfield array and insert the flatfield data.

    PPSFlatArrayH = array(PPSFlatDataSetH, "PRIMARY", modify)
    PPSFlatP => real32Array2Data(PPSFlatArrayH)

    if (goodFlatField) then

        PPSFlatP = 0.0
  
        do i = xMin, xMax
            x = min(imax, floor((i - xMin) / realBinX))
            do j = yMin, yMax
                y = min(floor((j - yMin) / realBinY), jMax)
     
                PPSFlatP(x, y) = PPSFlatP(x, y) + shiftedFlat(i, j) / binXTimesBinY
         
            end do
        end do
    else
        PPSFlatP = 1.0
    end if
    call release(PPSFlatArrayH)
    call release(PPSFlatDataSetH)
     
end subroutine  CreateFlatFieldedImageFile
!***********************************************************
! Subroutine SetQualityArray
! Sets the quality array flag
!***********************************************************
subroutine SetQualityArray(xMin, xMax, yMin, yMax)

    integer(kind=int32), intent(in) :: xMin                                       ! Minimum x value for window.
    integer(kind=int32), intent(in) :: xMax                                       ! Maximum x value for window.
    integer(kind=int32), intent(in) :: yMin                                       ! Minimum y value for window.
    integer(kind=int32), intent(in) :: yMax                                       ! Maximum y value for window.
    integer :: i, j
    integer :: ind, iMinusXminOverBinX

    do i = xMin, xMax
        iMinusXminOverBinX = int((real(i) - realXmin) / realBinX + 0.5) - 1
        do j = yMin, yMax

            if(abs(outOSW(i, j) - missingPixelValue) .lt. 0.5) then

                !*************************************
                ! This will create problems later in OMDETECT - 
                ! so flag the quality array so OMDETECT can take
                ! appropriate action.
                !
                ! Map the (i,j) element to the corresponding 
                ! element in quality array
                !******************************************  
                write(errstr,*)'Flagging quality array at pixel ', &                              
                     int((real(j) - realYMin) / realBinY + 0.5)
                call message(AppMsg, Verbose)

                !****************************************************
                ! Set the quality array element in this case to 16 - 
                ! implies unknown reason
                ! This is incorrect here as this corresponds     
                !****************************************************
                ind =  int((real(j) - realYMin) / realBinY + 0.5) - 1
                qualArrayP(iMinusXMinOverBinX, ind) = 16            
            end if
        end do
    end do
end subroutine SetQualityArray
!*******************************************
! Subroutine CreateShiftedFlat
!*******************************************
subroutine CreateShiftedFlat(inOrbitFlatArrayH, nFrames, &
               windowX0, windowDX,  windowY0, windowDY)
    type(ArrayT), intent(in) :: inOrbitFlatArrayH    ! In Orbit Flatfield array handle
    integer(kind=int16), intent(in) :: nFrames       ! Number of tracking frames
    integer(kind=int32), intent(in) :: windowX0      ! WindowX0
    integer(kind=int32), intent(in) :: windowY0      ! WindowY0
    integer(kind=int32), intent(in) :: windowDX      ! WindowDX
    integer(kind=int32), intent(in) :: windowDY      ! WindowDY
  
    !**************************************
    ! Local variables
    !**************************************
    real(kind=single), allocatable, dimension(:, :) :: inOrbitFlat    ! In orbit flatfield.
    real(kind=single), allocatable, dimension(:, :) :: oShiftedFlat   ! Oversampled tracking shifted flatfield.
    real(kind=single), allocatable, dimension(:, :) :: lInOrbitFlat   ! Localised in orbit flatfield.
    real(kind=single), allocatable, dimension(:, :) :: oInOrbitFlat   ! Oversampled in orbit flatfield.
    real(kind=single) :: rShiftX                                      ! Relative shift in x.
    real(kind=single) :: rShiftY                                      ! Relative shift in y.
    real(kind=single), dimension(0 : 10)            :: jOverTen = &
               (/ 0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0/)     

    integer :: allocErr
    integer :: i, j, x, y
    integer(kind=int32) :: maxDX                                      ! Maximum drift in x (centroided pixels).
    integer(kind=int32) :: maxDY                                      ! Maximum drift in y (centroided pixels).
    integer(kind=int32) :: cpxMin                                     ! Minimum x value for window (centroided pixels).
    integer(kind=int32) :: cpxMax                                     ! Maximum x value for window (centroided pixels).
    integer(kind=int32) :: cpyMin                                     ! Minimum y value for window (centroided pixels).
    integer(kind=int32) :: cpyMax                                     ! Maximum y value for window (centroided pixels).
    integer(kind=int32) :: opxMin                                     ! Minimum x value for window (oversampled pixels).
    integer(kind=int32) :: opxMax                                     ! Maximum x value for window (oversampled pixels).
    integer(kind=int32) :: opyMin                                     ! Minimum y value for window (oversampled pixels).
    integer(kind=int32) :: opyMax                                     ! Maximum y value for window (oversampled pixels).
    integer(kind=int32) :: oMaxDX                                     ! Maximum drift in x (oversampled pixels).
    integer(kind=int32) :: oMaxDY                                     ! Maximum drift in y (oversampled pixels).
    integer(kind=int32) :: iShiftX                                    ! Incremental shift in x.
    integer(kind=int32) :: iShiftY                                    ! Incremental shift in y.
    integer(kind=int32)                             :: ishiftX0
    integer(kind=int32)                             :: ishiftY0
    integer(kind=int32)                             :: iMinus1       ! stores i-1 in loops
     
  
  
    ! PSW Assume in orbit flat fields are BPE binned and not DPU binned.

    ! Alloate space for the in orbit flatfield.

    allocate (inOrbitFlat(0 : 2047, 0 : 2047), stat = allocErr)
    call CheckAllocation(allocErr, "inOrbitFlat")


    ! Fill the in orbit flat field.

    inOrbitFlat = 0.0
    
    do i = 0, 2047
        x = floor(i / 2.0)
        do j = 0, 2047
            y = floor(j / 2.0)
            inOrbitFlat(i, j) = inOrbitFlatP(x, y)
        end do
    end do


    ! Release handles on data.

    call release(inOrbitFlatArrayH)
  
    !*************************************************
    ! Calculate the maximum drift in x and y.
    ! We add 0.5 to the drifts so that positions 
    ! are calculated with respect to the centre of the
    ! pixel rather than the bottom left hand corner.
    !**************************************************

    maxDX = 0
    maxDY = 0
    do i = 1, nFrames
        maxDX = max(maxDX, ceiling(0.5 + abs(tf(i) % dX)))
        maxDY = max(maxDY, ceiling(0.5 + abs(tf(i) % dY)))
    end do

    ! Allocate space for the localised in orbit flatfield.
    cpxMin = windowX0 - maxDx
    cpxMax = windowX0 + windowDX - 1 + maxDX
    cpyMin = windowY0 - maxDY
    cpyMax = windowY0 + windowDY - 1 + maxDY
    allocate (lInOrbitFlat(cpxMin : cpxMax, cpyMin : cpyMax), stat=allocErr)
    call CheckAllocation(allocErr, "lInOrbitFlat")


    !****************************************
    ! Fill the localised in orbit flatfield.
    !****************************************

    lInOrbitFlat = 0.0
    do i = max0(0, cpxMin), min0(2047, cpxMax)    
        do j = max0(0, cpyMin), min0(2047, cpyMax)
            lInOrbitFlat(i, j) = inOrbitFlat(i, j)
        end do
    end do

    ! Free up memory.
    deallocate(inOrbitFlat, stat=allocErr)
    call CheckDeallocation(allocErr, "inOrbitFlat")


    ! Allocate space for the oversampled in orbit flatfield.

    oMaxDX = maxDX * sampleFactor
    oMaxDY = maxDY * sampleFactor
    opxMin = 0 - oMaxDX
    opxMax = ((cpxMax - cpxMin - (2 * maxDX)) * sampleFactor) + oMaxDX
    opyMin = 0 - oMaxDY
    opyMax = ((cpyMax - cpyMin - (2 * maxDY)) * sampleFactor) + oMaxDY

    allocate (oInOrbitFlat(opxMin : opxMax, opyMin : opyMax), stat=allocErr)
    call CheckAllocation(allocErr, "oInOrbitFlat")


    !*********************************************
    ! Oversample the localised in orbit flatfield.
    !*********************************************

    oInOrbitFlat = 0.0
    do i = opxMin, opxMax
        x = cpxMin + floor((i + oMaxDX) / realSampleFactor)
        do j = opyMin, opyMax
            y = cpyMin + floor((j + oMaxDY) / realSampleFactor)
  
            oInOrbitFlat(i, j) = lInOrbitFlat(x, y)
        end do
    end do

    ! Free up memory.

    deallocate(lInOrbitFlat, stat=allocErr)
    call CheckDeallocation(allocErr, "lInOrbitFlat")


    ! Allocate space for the oversampled tracking shifted flatfield.

    opxMin = 0
    opxMax = (cpxMax - cpxMin - (2 * maxDX)) * sampleFactor   !cpxmin + crashes!
    opyMin = 0
    opyMax = (cpyMax - cpyMin - (2 * maxDY)) * sampleFactor   ! cpymin+ crashes!


    allocate (oShiftedFlat(opxMin : opxMax, opyMin : opyMax), stat=allocErr)
    call CheckAllocation(allocErr, "oShiftedFlat")
    
    !**************************************************************************************
    ! Loop over the tracking frames and construct an oversampled tracking shifted flatfield.
    ! When we apply the shifts, we must centre the star in the pixel, so we add 0.5 to x
    ! and y or in this case to dx and dy.
    !*************************************************************************************


    oShiftedFlat = 0.0
    do i = 1, nFrames
        iMinus1 = i - 1
        rShiftX = (tf(i) % dx - tf(iMinus1) % dx) * sampleFactor
        rShiftY = (tf(i) % dy - tf(iMinus1) % dy) * sampleFactor
        iShiftX0 = ((0.5 + tf(iMinus1) % dx) * sampleFactor) 
        iShiftY0 = ((0.5 + tf(iMinus1) % dy) * sampleFactor) 


        do j = 0, 10

            iShiftX = nint(ishiftX0 + (rShiftX * jOverTen(j)))  
            iShiftY = nint(ishiftY0 + (rShiftY * jOverTen(j)))  


            oShiftedFlat = oShiftedFlat +   & 
                           oInOrbitFlat(opxMin - iShiftX : opxMax - iShiftX, &
                                        opyMin - iShiftY : opyMax - iShiftY) / 11.0
        end do
    end do

    !**************************************
    ! Normalise the summed flatfield.
    !************************************** 

    if (nFrames .eq. 0) then
        oShiftedFlat = oInOrbitFlat     ! Changed on 02.09.20001  was  0.5  
    else
        oShiftedFlat = oShiftedFlat / real(nFrames)
    end if  


    deallocate(oInOrbitFlat, stat=allocErr)
    call CheckDeallocation(allocErr, "oInOrbitFlat")



    cpxMin = windowX0
    cpxMax = windowX0 + windowDX - 1
    cpyMin = windowY0
    cpyMax = windowY0 + windowDY - 1
    allocate (shiftedFlat(cpxMin : cpxMax, cpyMin : cpyMax), stat=allocErr)
    call CheckAllocation(allocErr, "shiftedFlat")


    !*****************************************************
    ! Create the tracking shifted flatfield.
    !*****************************************************

    shiftedFlat = 0.0
    opxMin = 0
    opxMax = (cpxMax - cpxMin) * sampleFactor
    opyMin = 0
    opyMax = (cpyMax - cpyMin) * sampleFactor


    do i = opxMin, opxMax
        x = cpxMin + floor(i / realSampleFactor)
        do j = opyMin, opyMax
            y = cpyMin + floor(j / realSampleFactor)
            shiftedFlat(x, y) = shiftedFlat(x, y) + oShiftedFlat(i, j) / realSampleFactor     
        end do
    end do

    if (minval(shiftedFlat) .le. 0.0) then
            call warning("zeroValuedFlat", "One or more elements of the flat-field array were" &
                    // " found to be zero. Resetting value to unity.")
    end if    

    ! Free up memory.
    deallocate(oShiftedFlat, stat = allocErr)
    call CheckDeallocation(allocErr, "oShiftedFlat")


end subroutine CreateShiftedFlat
end module omflatfield_mod








